(* :Title: SavitzkyGolay *)

(* :Author: Arnd Roth
	Abteilung Zellphysiologie
	Max-Planck-Institut fuer Medizinische Forschung
	Postfach 10 38 20, D-69028 Heidelberg, Germany
	email: roth@sunny.mpimf-heidelberg.mpg.de *)

(* :Summary: This little package implements Savitzky-Golay smoothing filters.
	These FIR lowpass filters can be thought of as a generalized moving
	average. Their coefficients are chosen such as to preserve higher
	moments in the data, thus reducing the distortion of essential features
	of the data like peak heights and line widths in a spectrum, while
	the efficiency of the suppression of random noise is effectively
	unchanged. See Source for a discussion of the details. *)

(* :Package Version: 1.0 *)

(* :Copyright: Copyright 1994, Max Planck Institute for Medical Research *)

(* :Context: Statistics`SavitzkyGolay` *)

(* :History: Version 1.0 by Arnd Roth, June, 1994. *)

(* :Keywords: data smoothing, FIR lowpass filter, perservation of higher
	moments of the data, least-squares fit of a polynomial, spectroscopy *)

(* :Source: William H. Press, Saul A. Teukolsky, William T. Vetterling and
	Brian P. Flannery, Numerical Recipes in C, Second Edition, Cambridge
	University Press 1992, Chapter 14.8 and references therein. *)

(* :Warning: none *)

(* :Mathematica Version: 2.1 *)

(* :Limitation: SavitzkyGolayFilter[data, coefficients] is restricted to data
	taken at a constant sampling interval. *)

(* :Discussion: see Source. *)

(* set up the package context *)

BeginPackage["Statistics`SavitzkyGolay`"]

(* usage messages for the exported functions and the context itself *)

SavitzkyGolay::usage = "SavitzkyGolay.m is a package that implements
	Savitzky-Golay smoothing filters. These FIR lowpass filters can be
	thought of as a generalized moving average. Their coefficients are
	chosen such as to preserve higher moments in the data, thus reducing
	the distortion of essential features of the data like peak heights
	and line widths in a spectrum, while the efficiency of the
	suppression of random noise is effectively unchanged."

SavitzkyGolayCoefficients::usage = "SavitzkyGolayCoefficients[nleft, nright,
	order] gives a list of coefficients {c[-nleft], c[-nleft + 1], ...,
	c[nright]} of a Savitzky-Golay FIR lowpass filter. The output of the
	filter, g[i], is a linear combination of input function values f[i]
	according to g[i] = Sum[c[n] f[i+n], {n, -nleft, nright}]. Typical
	values for the order are 2 and 4."

SavitzkyGolayFilter::usage = "SavitzkyGolayFilter[data, coefficients] applies
	a Savitzky-Golay filter defined by the list of coefficients to a data
	vector."

Begin["`Private`"]

(* error messages for exported objects *)

SavitzkyGolay::hord = "The requested order `1` is too high."

SavitzkyGolay::norm = "Warning: the filter coefficients sum up to `1`."


(* definition of the exported functions *)

SavitzkyGolayCoefficients[(nleft_Integer)?Positive,
                          (nright_Integer)?Positive,
                          (order_Integer)?Positive] :=
Module[{a, aTaInverse},
a = Table[If[i == j == 0, 1, i^j], {i, -nleft, nright}, {j, 0, order}];
aTaInverse = Inverse[Transpose[a] . a];
Table[First[aTaInverse .
            (Transpose[a] . Table[If[i == n, 1, 0], {i, -nleft, nright}])
      ], {n, -nleft, nright}]
] /; nleft + nright >= order || Message[SavitzkyGolay::hord, order]

SavitzkyGolayFilter[data_?VectorQ, coefficients_?VectorQ] :=
Map[coefficients . # &, Partition[data, Length[coefficients], 1]] /;
Plus @@ coefficients == 1 ||
Message[SavitzkyGolay::norm, Plus @@ coefficients] || True


End[]   (* end the private context *)

Protect[SavitzkyGolayCoefficients, SavitzkyGolayFilter]
(* protect exported symbols *)

EndPackage[]   (* end the package context *)

