/**
 *  @file      dataTypes.h
 *  @brief     Number datatypes Int and Real.
 *  @author    Mikica Kocic
 *  @copyright GNU General Public License (GPLv3).
 */

/////////////////////////////////////////////////////////////////////////////////////////
/** @defgroup g1 Number datatypes                                                      */
/////////////////////////////////////////////////////////////////////////////////////////
                                                                                   /*@{*/
typedef long Int;

/////////////////////////////////////////////////////////////////////////////////////////

#ifdef USE_QUADMATH // 128-bit floating point

    #include <quadmath.h>

    typedef __float128  Real;

    #define strToReal   strtoflt128
    #define Power(v,e)  powq(v,e)
    #define Sqrt(v)     sqrtq(v)
    #define sqrt(v)     sqrtq(v)
    #define round(v)    roundq(v)

    /** @warning Without std::, abs() is a GCC built-in function, which returns an int.
     */
    inline static Real abs( Real v ) { return fabsq(v); }

    inline static std::ostream& operator<< ( std::ostream& os, const Real& value ) {
        char str[256];
        quadmath_snprintf( str, sizeof(str), "%Qg", value );
        return os << str;
    }

    inline static void fputReal( FILE* outf, Real value ) {
        char str[256];
        quadmath_snprintf( str, 256, "%Qf", value );
        fputs( str, outf );
    }

    inline static Real operator ""_q ( const char* str ) {
        return strtoflt128( str, NULL );
    }

#else
    typedef double      Real;

    #define strToReal   strtod
    #define Power(v,e)  pow(v,e)
    #define Sqrt(v)     sqrt(v)
    #define Erf(v)      erf(v)
    #define Tanh(v)     tanh(v)

    /** @warning Without std::, abs() is a GCC built-in function, which returns an int.
     */
    inline static Real abs( Real v ) { return std::fabs(v); }

    inline static void fputReal( FILE* outf, Real value ) {
        fprintf( outf, "%lf", value );
    }
#endif


const Real TINY_Real = 1e-100; // std::numeric_limits<Real>::min();


/** Reads an array of real numbers from a string.
 */
static size_t sscanf_Real( const char* line, Real* data, size_t count )
{
    for( size_t i = 0; i < count; ++ i )
    {
        char* endPtr = NULL;
        data[i] = strToReal( line, &endPtr );
        if( endPtr == line ) return i;
        line = endPtr;
    }

    return count;
}

// static const Real PI = acos(-1.0);

static inline Real pow2 ( Real v ) { return v * v; }
static inline Real pow3 ( Real v ) { return v * v * v; }
static inline Real pow15( Real v ) { return pow( v, 1.5 ); }
                                                                                   /*@}*/
/////////////////////////////////////////////////////////////////////////////////////////
